import asyncio
import io
import logging
from base64 import b64decode
from dataclasses import dataclass
from typing import AsyncIterator
from uuid import uuid4

import aiohttp
import tornado
from PIL import Image
from tornado.escape import json_decode, json_encode
from tornado.httpclient import HTTPRequest
from tornado.ioloop import PeriodicCallback
from tornado.websocket import WebSocketClientConnection, websocket_connect

logger = logging.getLogger(__name__)

# matplotlib 설정 상수
# 그래프의 기본 크기 (가로, 세로) - 인치 단위
# 10x6 인치는 웹 환경에서 보기 좋은 와이드스크린 비율입니다
FIGURE_SIZE = (10, 6)

# 그래프 저장 시 기본 파일 형식
# 'svg': 벡터 그래픽 형식으로 확대해도 품질이 저하되지 않음
# 다른 옵션: 'png', 'pdf', 'jpg' 등
SAVEFIG_FORMAT = "svg"

# IPython에서 표시할 그래프 형식 목록
# 여러 형식을 지정하면 모두 생성되어 클라이언트에서 선택 가능
# 'svg': 벡터 그래픽으로 확대/축소 및 동적 조작에 적합
# 'png': 래스터 이미지로 모든 환경에서 일관된 표시 가능
MATPLOTLIB_FORMATS = ["svg", "png"]


class ConnectionError(Exception):
    """Exception raised when connection to an IPython kernel fails."""

    pass


class ExecutionError(Exception):
    """Exception raised when code execution in the IPython kernel fails.

    Args:
        message: Error message
        trace: Stack trace string representation
    """

    def __init__(self, message: str, trace: str | None = None):
        super().__init__(message)
        self.trace = trace


@dataclass
class ExecutionResult:
    """The result of a code execution.

    Args:
        text: Output text generated during execution
        images: List of images generated during execution
        svg_images: List of SVG XML strings generated during execution
    """

    text: str | None
    images: list[Image.Image]
    svg_images: list[str]


class Execution:
    """Represents a code execution in an IPython kernel.

    Args:
        client: The client instance that created this execution
        req_id: Unique identifier for the execution request
    """

    def __init__(self, client: "ExecutionClient", req_id: str):
        self.client = client
        self.req_id = req_id

        self._chunks: list[str] = []
        self._images: list[Image.Image] = []
        self._svg_images: list[str] = []

        self._stream_consumed: bool = False

    async def result(self, timeout: float = 120) -> ExecutionResult:
        """Waits for execution to complete and returns the final result.

        If a timeout is reached, the kernel is interrupted.

        Args:
            timeout: Maximum time to wait in seconds. Defaults to 120.

        Returns:
            ExecutionResult object

        Raises:
            asyncio.TimeoutError: If execution exceeds timeout duration
        """
        if not self._stream_consumed:
            async for _ in self.stream(timeout=timeout):
                pass

        return ExecutionResult(
            text="".join(self._chunks).strip() if self._chunks else None,
            images=self._images,
            svg_images=self._svg_images,
        )

    async def stream(self, timeout: float = 120) -> AsyncIterator[str]:
        """Streams the execution output text as it becomes available.

        Args:
            timeout: Maximum time to wait in seconds. Defaults to 120.

        Yields:
            Output text chunks as they arrive

        Raises:
            asyncio.TimeoutError: If execution exceeds timeout duration
        """
        try:
            async with asyncio.timeout(timeout):
                async for elem in self._stream():
                    match elem:
                        case str():
                            self._chunks.append(elem)
                            yield elem
                        case Image.Image():
                            self._images.append(elem)
        except asyncio.TimeoutError:
            await self.client._interrupt_kernel()
            await asyncio.sleep(0.2)  # TODO: make configurable
            raise
        finally:
            self._stream_consumed = True

    async def _stream(self) -> AsyncIterator[str | Image.Image]:
        while True:
            msg_dict = await self.client._read_message()
            msg_type = msg_dict["msg_type"]
            msg_id = msg_dict["parent_header"].get("msg_id", None)

            if msg_id != self.req_id:
                continue

            if msg_type == "stream":
                yield msg_dict["content"]["text"]
            elif msg_type == "error":
                self._raise_error(msg_dict)
            elif msg_type == "execute_reply":
                if msg_dict["content"]["status"] == "error":
                    self._raise_error(msg_dict)
                break
            elif msg_type in ["execute_result", "display_data"]:
                msg_data = msg_dict["content"]["data"]
                yield msg_data["text/plain"]

                # SVG 형식 처리
                if "image/svg+xml" in msg_data:
                    svg_data = msg_data["image/svg+xml"]
                    self._svg_images.append(svg_data)
                    yield svg_data

                # PNG 형식 처리
                if "image/png" in msg_data:
                    image_bytes_io = io.BytesIO(b64decode(msg_data["image/png"]))
                    image = Image.open(image_bytes_io)
                    image.load()
                    yield image

    def _raise_error(self, msg_dict):
        error_name = msg_dict["content"].get("ename", "Unknown Error")
        error_value = msg_dict["content"].get("evalue", "")
        error_trace = "\n".join(msg_dict["content"]["traceback"])
        raise ExecutionError(f"{error_name}: {error_value}", error_trace)


class ExecutionClient:
    """A context manager for executing code in an IPython kernel.

    Args:
        host: Hostname where the code execution container is running
        port: Host port of the code execution container
        heartbeat_interval: Interval in seconds between heartbeat pings. Defaults to 10.

    Example:
        ```python
        from jarvis_ipybox import ExecutionClient, ExecutionContainer

        binds = {"/host/path": "example/path"}
        env = {"API_KEY": "secret"}

        async with ExecutionContainer(binds=binds, env=env) as container:
            async with ExecutionClient(host="localhost", port=container.port) as client:
                result = await client.execute("print('Hello, world!')")
                print(result.text)
        ```
        > Hello, world!
    """

    def __init__(
        self,
        port: int,
        host: str = "localhost",
        heartbeat_interval: float = 10,
        https: bool = False,
        kernel_id: str = None
    ):
        self.port = port
        self.host = host
        self.https = https

        self._heartbeat_interval = heartbeat_interval
        self._heartbeat_callback = None

        self._kernel_id = kernel_id
        self._ws: WebSocketClientConnection

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()

    @property
    def kernel_id(self):
        """The ID of the running IPython kernel.

        Raises:
            ValueError: If not connected to a kernel
        """
        if self._kernel_id is None:
            raise ValueError("Not connected to kernel")
        return self._kernel_id

    @property
    def base_http_url(self):
        if self.https:
            return f"https://{self.host}:{self.port}/api/kernels"
        else:
            return f"http://{self.host}:{self.port}/api/kernels"

    @property
    def kernel_http_url(self):
        return f"{self.base_http_url}/{self.kernel_id}"

    @property
    def kernel_ws_url(self):
        if self.https:
            return f"wss://{self.host}:{self.port}/api/kernels/{self.kernel_id}/channels"
        else:
            return f"ws://{self.host}:{self.port}/api/kernels/{self.kernel_id}/channels"

    async def connect(self, retries: int = 10, retry_interval: float = 1.0):
        """Creates and connects to an IPython kernel.

        Args:
            retries: Number of connection attempts. Defaults to 10.
            retry_interval: Delay between retries in seconds. Defaults to 1.0.

        Raises:
            ConnectionError: If connection cannot be established after all retries
        """
        if self._kernel_id is None:
            for _ in range(retries):
                    try:
                        self._kernel_id = await self._create_kernel()
                        break
                    except Exception:
                        await asyncio.sleep(retry_interval)
            else:
                raise ConnectionError("Failed to create kernel")

        self._ws = await websocket_connect(HTTPRequest(url=self.kernel_ws_url))
        logger.info("Connected to kernel")

        self.heartbeat_callback = PeriodicCallback(
            self._ping_kernel, self._heartbeat_interval * 1000
        )
        self.heartbeat_callback.start()
        logger.info(f"Started heartbeat (interval = {self._heartbeat_interval}s)")

        await self._init_kernel()

    async def disconnect(self):
        """Closes the connection to the kernel and cleans up resources."""
        self.heartbeat_callback.stop()
        self._ws.close()
        # async with aiohttp.ClientSession() as session:
        #     async with session.delete(self.kernel_http_url):
        #         pass

    async def execute(self, code: str, timeout: float = 120) -> ExecutionResult:
        """Executes code and returns the result.

        Args:
            code: Code to execute
            timeout: Maximum execution time in seconds. Defaults to 120.

        Returns:
            ExecutionResult object

        Raises:
            ExecutionError: If code execution raised an error
            asyncio.TimeoutError: If execution exceeds timeout duration
        """
        execution = await self.submit(code)
        return await execution.result(timeout=timeout)

    async def submit(self, code: str) -> Execution:
        """Submits code for execution and returns an Execution object to track it.

        Args:
            code: Python code to execute

        Returns:
            An Execution object to track the submitted code execution
        """
        req_id = uuid4().hex
        req = {
            "header": {
                "username": "",
                "version": "5.0",
                "session": "",
                "msg_id": req_id,
                "msg_type": "execute_request",
            },
            "parent_header": {},
            "channel": "shell",
            "content": {
                "code": code,
                "silent": False,
                "store_history": False,
                "user_expressions": {},
                "allow_stdin": False,
            },
            "metadata": {},
            "buffers": {},
        }

        await self._send_request(req)
        return Execution(client=self, req_id=req_id)

    async def get_module_sources(self, module_names: list[str]) -> str | None:
        result = await self.execute(f"print_module_sources({module_names})")
        return result.text

    async def _send_request(self, req):
        await self._ws.write_message(json_encode(req))

    async def _read_message(self) -> dict:
        return json_decode(await self._ws.read_message())

    async def _create_kernel(self):
        async with aiohttp.ClientSession() as session:
            async with session.post(url=self.base_http_url, json={"name": "python"}) as response:
                kernel = await response.json()
                return kernel["id"]

    async def _interrupt_kernel(self):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.kernel_http_url}/interrupt", json={"kernel_id": self._kernel_id}
            ) as response:
                logger.info(f"Kernel interrupted: {response.status}")

    async def _ping_kernel(self):
        try:
            self._ws.ping()
        except tornado.iostream.StreamClosedError as e:
            logger.error("Kernel disconnected", e)

    async def _init_kernel(self):
        await self.execute(
            r"%colors nocolor"
            + "\n# 필요한 모듈 임포트"
            + "\nfrom jarvis_ipybox.modinfo import print_module_sources, get_module_info"
            + "\nimport matplotlib"
            + "\nimport matplotlib.pyplot as plt"
            + "\nfrom matplotlib_inline.backend_inline import set_matplotlib_formats"
            + "\n# matplotlib 설정"
            # + "\nmatplotlib.use('Agg')"  # 백엔드 설정
            + f"\nplt.rcParams['figure.figsize'] = {FIGURE_SIZE}"
            + f"\nplt.rcParams['savefig.format'] = '{SAVEFIG_FORMAT}'"
            + f"\nset_matplotlib_formats({', '.join(repr(fmt) for fmt in MATPLOTLIB_FORMATS)})"
        )



import asyncio
from collections import defaultdict
from pathlib import Path

from aiodocker import Docker
from aiodocker.containers import DockerContainer

from jarvis_ipybox.utils import arun

DEFAULT_TAG = "jarvis-ipybox"


class ExecutionContainer:
    """
    A context manager for managing the lifecycle of a Docker container used for code execution.

    It handles the creation, port mapping, volume binding, and cleanup of the container.

    Args:
        tag: Tag of the Docker image to use (defaults to jarvis-ipybox)
        binds: Mapping of host paths to container paths for volume mounting.
            Host paths may be relative or absolute. Container paths must be relative
            and are created as subdirectories of `/app` in the container.
        env: Environment variables to set in the container
        port: Host port to map to the container's executor port. If not provided,
            a random port will be allocated.
        port_allocation_timeout: Timeout in seconds waiting for the container's host
            port to be allocated. This is only relevant on OSX when `port=None`.
        show_pull_progress: Whether to show progress when pulling the Docker image.

    Attributes:
        port: Host port mapped to the container's executor port. This port is dynamically
            allocated when the container is started.

    Example:
        ```python
        from jarvis_ipybox import ExecutionClient, ExecutionContainer

        binds = {"/host/path": "example/path"}
        env = {"API_KEY": "secret"}

        async with ExecutionContainer(binds=binds, env=env) as container:
            async with ExecutionClient(host="localhost", port=container.port) as client:
                result = await client.execute("print('Hello, world!')")
                print(result.text)
        ```
        > Hello, world!
    """

    def __init__(
        self,
        tag: str = DEFAULT_TAG,
        binds: dict[str, str] | None = None,
        env: dict[str, str] | None = None,
        port: int | None = None,
        port_allocation_timeout: float = 10,
        show_pull_progress: bool = True,
    ):
        self.tag = tag
        self.binds = binds or {}
        self.env = env or {}
        self.show_pull_progress = show_pull_progress

        self._docker = None
        self._container = None
        self._port = port
        self._port_allocation_timeout = port_allocation_timeout
        self._executor_port = 8888

    async def __aenter__(self):
        try:
            await self.run()
        except Exception as e:
            await self.kill()
            raise e
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.kill()

    @property
    def port(self) -> int:
        """
        The host port mapped to the container's executor port.

        This port is dynamically allocated when the container is started unless
        explicitly provided.

        Raises:
            RuntimeError: If the container is not running
        """
        if self._port is None:
            raise RuntimeError("Container not running")
        return self._port

    async def kill(self):
        """
        Kill and remove the Docker container.
        """
        if self._container:
            await self._container.kill()

        if self._docker:
            await self._docker.close()

    async def run(self):
        """
        Create and start the Docker container.
        """
        self._docker = Docker()
        self._container = await self._run(self._executor_port)

    async def _run(self, executor_port: int = 8888):
        host_port = {"HostPort": str(self._port)} if self._port else {}
        executor_port_key = f"{executor_port}/tcp"

        config = {
            "Image": self.tag,
            "HostConfig": {
                "PortBindings": {executor_port_key: [host_port]},
                "AutoRemove": True,
                "Binds": await self._container_binds(),
            },
            "Env": self._container_env(),
            "ExposedPorts": {executor_port_key: {}},
        }

        if not await self._local_image():
            await self._pull_image()

        container = await self._docker.containers.create(config=config)  # type: ignore
        await container.start()

        self._port = await self._host_port(container, executor_port_key)
        return container

    async def _host_port(self, container: DockerContainer, executor_port_key: str) -> int:
        try:
            async with asyncio.timeout(self._port_allocation_timeout):
                while True:
                    container_info = await container.show()
                    host_ports = container_info["NetworkSettings"]["Ports"].get(executor_port_key)
                    if host_ports and host_ports[0].get("HostPort"):
                        return int(host_ports[0]["HostPort"])
                    await asyncio.sleep(0.1)

        except TimeoutError:
            _timeout = self._port_allocation_timeout
            raise TimeoutError(
                f"Timed out waiting for host port allocation after {_timeout} seconds"
            )

    async def _local_image(self) -> bool:
        tag = self.tag if ":" in self.tag else f"{self.tag}:latest"

        images = await self._docker.images.list()  # type: ignore
        for img in images:
            if "RepoTags" in img and img["RepoTags"] is not None and tag in img["RepoTags"]:
                return True

        return False

    async def _pull_image(self):
        # Track progress by layer ID
        layer_progress = defaultdict(str)

        async for message in self._docker.images.pull(self.tag, stream=True):  # type: ignore
            if not self.show_pull_progress:
                continue

            if "status" in message:
                status = message["status"]
                if "id" in message:
                    layer_id = message["id"]
                    if "progress" in message:
                        layer_progress[layer_id] = f"{status}: {message['progress']}"
                    else:
                        layer_progress[layer_id] = status

                    # Clear screen and move cursor to top
                    print("\033[2J\033[H", end="")
                    # Print all layer progress
                    for layer_id, progress in layer_progress.items():
                        print(f"{layer_id}: {progress}")
                else:
                    # Status without layer ID (like "Downloading" or "Complete")
                    print(f"\r{status}", end="")

        if self.show_pull_progress:
            print()

    async def _container_binds(self) -> list[str]:
        container_binds = []
        for host_path, container_path in self.binds.items():
            host_path_resolved = await arun(self._prepare_host_path, host_path)
            container_binds.append(f"{host_path_resolved}:/app/{container_path}")
        return container_binds

    def _prepare_host_path(self, host_path: str) -> Path:
        resolved = Path(host_path).resolve()
        if not resolved.exists():
            resolved.mkdir(parents=True)
        return resolved

    def _container_env(self) -> list[str]:
        return [f"{k}={v}" for k, v in self.env.items()]



#!/usr/bin/env python3
import os

from jupyter_client.kernelspec import KernelSpecManager
from jupyter_server.services.kernels.kernelmanager import MappingKernelManager


class CustomMappingKernelManager(MappingKernelManager):
    """
    각 커널이 독립된 디렉토리를 사용하도록 설정하는 커스텀 커널 매니저
    """
    
    def __init__(self, **kwargs):
        
        # 커널 유효 시간
        kwargs['cull_idle_timeout'] = 300  # 5분
        kwargs['cull_interval'] = 60        # 1분마다 검사
        kwargs['cull_connected'] = False    # 연결된 커널도 종료

        super().__init__(**kwargs)
        # root_dir 속성 직접 설정
        self.root_dir = os.environ.get('HOME', '/home/appuser')
    
    def _kernel_spec_manager_default(self):
        return KernelSpecManager()
    
    def start_kernel(self, kernel_id=None, path=None, **kwargs):
        """
        커널 시작 시 독립된 디렉토리를 설정합니다.
        """
        kernel_id = kernel_id or self.new_kernel_id()
        
        # 커널별 독립 디렉토리 설정
        kernel_dir = self._get_kernel_dir(kernel_id)
        os.environ['JUPYTER_DATA_DIR'] = kernel_dir
        
        # 환경 변수 설정
        env = kwargs.pop('env', {})
        env['KERNEL_ID'] = kernel_id
        env['KERNEL_DIR'] = kernel_dir
        
        # PYTHONPATH에 /app 디렉토리 추가
        python_path = env.get('PYTHONPATH', '')
        if python_path:
            env['PYTHONPATH'] = f"/app:{python_path}"
        else:
            env['PYTHONPATH'] = "/app"
        
        kwargs['env'] = env
        
        # 작업 디렉토리를 커널 디렉토리로 설정
        if path is None:
            path = kernel_dir
            
        # cwd 직접 설정하여 cwd_for_path 호출 방지
        kwargs['cwd'] = kernel_dir
        
        return super().start_kernel(kernel_id=kernel_id, path=path, **kwargs)
    
    def cwd_for_path(self, path, env=None):
        """
        cwd_for_path 메서드 오버라이드하여 root_dir 참조 오류 방지
        """
        if path is None:
            return self._get_kernel_dir(env.get('KERNEL_ID', '')) if env else os.getcwd()
        
        return os.path.abspath(path)
    
    def _get_kernel_dir(self, kernel_id):
        """
        커널 ID에 해당하는 디렉토리 경로를 반환합니다.
        """
        home = os.environ.get('HOME', '/home/appuser')
        kernel_root = os.environ.get('KERNEL_ROOT_DIR', f'{home}/data')
        
        # 커널 루트 디렉토리가 없으면 생성
        if not os.path.exists(kernel_root):
            os.makedirs(kernel_root, exist_ok=True)
        
        # 커널 ID를 기반으로 디렉토리 생성
        kernel_dir = os.path.join(kernel_root, kernel_id)
        if not os.path.exists(kernel_dir):
            os.makedirs(kernel_dir, exist_ok=True)

        # 커널 ID별 인풋, 아웃풋 디렉토리 생성
        kernel_input_dir = os.path.join(kernel_root, kernel_id, "input")
        if not os.path.exists(kernel_input_dir):
            os.makedirs(kernel_input_dir, exist_ok=True)
        
        kernel_output_dir = os.path.join(kernel_root, kernel_id, "output")
        if not os.path.exists(kernel_output_dir):
            os.makedirs(kernel_output_dir, exist_ok=True)
            
        return kernel_dir


if __name__ == "__main__":
    # 테스트 코드 또는 필요한 초기화 코드
    manager = CustomMappingKernelManager()
    print("커스텀 커널 매니저가 초기화되었습니다.")


import importlib
from dataclasses import dataclass
from inspect import getfile, getsource
from pathlib import Path, PurePath, PurePosixPath
from typing import List


@dataclass
class ModuleInfo:
    name: str
    relpath: PurePath
    source: str


def print_module_sources(module_names: List[str]):
    module_info_strings = []
    for module_name in module_names:
        module_info = get_module_info(module_name)
        module_info_str = f"```python\n# Module: {module_info.name}\n\n{module_info.source}\n```"
        module_info_strings.append(module_info_str)
    print("\n\n".join(module_info_strings))


def get_module_info(module_name: str) -> ModuleInfo:
    module = importlib.import_module(module_name)
    module_path = Path(getfile(module))

    relscope = PurePosixPath(module_name.replace(".", "/"))

    if module_path.name == "__init__.py":
        relpath = relscope / "__init__.py"
    else:
        relpath = relscope.with_suffix(".py")

    try:
        code = getsource(module)
    except OSError:
        code = ""

    return ModuleInfo(name=module.__name__, relpath=relpath, source=code)



import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Annotated

import typer

from jarvis_ipybox.container import DEFAULT_TAG

pkg_path = Path(__file__).parent
app = typer.Typer()


@app.command()
def build(
    tag: Annotated[
        str,
        typer.Option(
            "--tag",
            "-t",
            help="Name and optionally a tag of the Docker image in 'name:tag' format",
        ),
    ] = DEFAULT_TAG,
    dependencies: Annotated[
        Path,
        typer.Option(
            "--dependencies",
            "-d",
            help="Path to dependencies file",
            exists=True,
            file_okay=True,
            dir_okay=False,
        ),
    ] = Path(__file__).parent
    / "config"
    / "default"
    / "dependencies.txt",
    root: Annotated[
        bool,
        typer.Option(
            "--root",
            "-r",
            help="Run container as root",
        ),
    ] = False,
):
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        with open(dependencies, "r") as f:
            dependencies_spec = f.read()

        with open(pkg_path / "config" / "default" / "pyproject.toml", "r") as f:
            project_spec = f.read()

        with open(tmp_path / "pyproject.toml", "w") as f:
            f.write(project_spec.format(dependencies=dependencies_spec))

        jarvis_ipybox_path = tmp_path / "jarvis_ipybox"
        jarvis_ipybox_path.mkdir()

        if root:
            dockerfile = "Dockerfile.root"
            build_cmd_args = []
        else:
            dockerfile = "Dockerfile"
            build_cmd_args = [
                "--build-arg",
                f"UID={os.getuid()}",
                "--build-arg",
                f"GID={os.getgid()}",
            ]

        shutil.copy(pkg_path / "modinfo.py", tmp_path / "jarvis_ipybox")
        shutil.copy(pkg_path / "config" / "default" / "environment.yml", tmp_path)
        shutil.copy(pkg_path / "config" / "default" / "custom_kernel_manager.py", tmp_path)
        shutil.copy(pkg_path / "docker" / dockerfile, tmp_path)
        shutil.copy(pkg_path / "scripts" / "server.sh", tmp_path)
        shutil.copy(pkg_path / "scripts" / "test.txt", tmp_path)

        build_cmd = [
            "docker",
            "build",
            "-f",
            tmp_path / dockerfile,
            "-t",
            tag,
            str(tmp_path),
            *build_cmd_args,
        ]

        process = subprocess.Popen(
            build_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )  # type: ignore

        while True:
            output = process.stdout.readline()  # type: ignore
            if output == "" and process.poll() is not None:
                break
            if output:
                print(output.strip())

        if process.returncode != 0:
            raise typer.Exit(code=1)


@app.command()
def cleanup(
    ancestor: Annotated[
        str,
        typer.Option(
            "--ancestor",
            "-a",
            help="Name and optionally a tag of the Docker ancestor image in 'name:tag' format",
        ),
    ] = DEFAULT_TAG,
):
    cleanup_script = pkg_path / "scripts" / "cleanup.sh"
    subprocess.run(["bash", str(cleanup_script), ancestor], capture_output=True, text=True)


if __name__ == "__main__":
    app()
